from gqla.abstracts import AbstractQuery, AbstractGenerator


class BasicQuery(AbstractQuery):
    def __init__(self, name=None, generator=None, item=None):
        super().__init__()
        self._body_raw = None
        self._generator = generator
        self._name = name
        self._item = item
        self._deprecated = False

    @property
    def body(self):
        return self._body

    @property
    def name(self):
        return self._name

    def set(self, name):
        self._name = name
        self._body = self.generator.generate(self._item)

    @property
    def generator(self):
        return self._generator

    @generator.setter
    def generator(self, value: AbstractGenerator):
        self._generator = value

    def args(self, args: list):
        self._args = args

    def regenerate(self, only_fields):
        self._body, self._deprecated = self.generator.generate(self._item, only_fields)

    @property
    def query(self):
        return self.name + (str(self._args) if len(self._args) > 0 else "") + self.body



from abc import ABC, abstractmethod


class AbstractExecutor(ABC):

    @abstractmethod
    def execute(self, *args, **kwargs):
        """Execute method"""


class AbstractRunner(ABC):

    @abstractmethod
    def run(self, *args):
        """Runner"""

    @abstractmethod
    def _can_query(self):
        """Checker"""



import asyncio
import json
import logging

import aiohttp
import requests
from aiohttp.client import DEFAULT_TIMEOUT

from gqla.abstracts import AbstractExecutor, AbstractRunner


class AsyncRunner(AbstractRunner):

    def __init__(self, properties=None):
        super().__init__()
        self._properties = properties

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, value):
        self._properties = value

    def _can_query(self):
        if self._properties.url is None:
            raise AttributeError

    async def run(self, pid, query):
        self._can_query()
        logging.info('Fetch async process {} started'.format(pid))
        async with aiohttp.request('POST', self._properties.url_string, json=query, timeout=DEFAULT_TIMEOUT) as resp:
            response = await resp.text()
        logging.info('Fetch async process {} ended'.format(pid))
        return json.loads(response)


class SyncRunner(AbstractRunner):

    def __init__(self, properties=None):
        super().__init__()
        self._properties = properties

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, value):
        self._properties = value

    def _can_query(self):
        if self._properties.url is None:
            raise AttributeError

    def run(self, pid, query):
        self._can_query()
        logging.info('Fetch sync process {} started'.format(pid))
        response = requests.post(self._properties.url_string, json=query, timeout=30)
        logging.info('Fetch async process {} ended'.format(pid))
        return json.loads(response.text)


class BasicExecutor(AbstractExecutor):

    def __init__(self, properties=None, storage=None, runner=None):
        super().__init__()
        self._properties = properties
        self._storage = storage
        self.runner = runner
        runner.properties = properties

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, value):
        self._properties = value
        self.runner.properties = value

    @property
    def storage(self):
        return self._storage.storage if self._storage is not None else None

    @storage.setter
    def storage(self, value):
        self._storage = value

    async def execute(self, pid='N/A', query=None, **kwargs):
        if self._storage is None:
            raise AttributeError
        if query is None:
            raise AttributeError
        if len(kwargs) > 0:
            params = "("
            for item in kwargs:
                params += ":".join([item, str(kwargs[item]).replace("'", '')])
            params += ")"
        else:
            params = ''
        if self.storage is not None:
            if str(query) in self.storage:
                instance = self.storage[query]
                instance.args(params)
                query = instance.query

                query = {
                    'query': self._properties.raw_query.format(query=query)
                }
        futures = [self.runner.run(pid, query=query)]
        done, pending = await asyncio.wait(futures)
        result = done.pop().result()
        return result



from abc import ABC, abstractmethod


class AbstractQuery(ABC):
    @abstractmethod
    def __init__(self):
        self._name = None
        self._body = None
        self._args = []

    @property
    @abstractmethod
    def body(self):
        """Define body"""

    @property
    @abstractmethod
    def name(self):
        """Define name"""

    @abstractmethod
    def set(self, *args):
        """Define setter"""

    @property
    @abstractmethod
    def generator(self):
        """Generator"""

    @property
    @abstractmethod
    def query(self):
        """Define getter"""



from gqla.GQQuery import BasicQuery
from gqla.GQQuery import BasicQueryGenerator, NormalRule, RecursiveRule
from gqla.abstracts import AbstractQuery, AbstractStorage


class BasicStorage(AbstractStorage):

    def __init__(self, properties=None, generator=None):
        super().__init__()
        self._query = {}
        self.generator = generator
        self._properties = properties
        self.generator.properties = properties

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, value):
        self._properties = value
        self.generator.properties = value

    @property
    def storage(self):
        return self._query

    @property
    def deprecates(self):
        deprecates = []
        for query in self._query:
            if self._query[query]._deprecated:
                deprecates.append(query)
        return deprecates

    def add(self, name: str, query: AbstractQuery):
        self._query[name] = query

    def create(self, alias, name, item, recursive_depth, only_fields=False):
        query = BasicQuery(name, self.generator, item)
        query.regenerate(only_fields)
        self.add(alias, query)



from gqla.abstracts import AbstractRule, AbstractGenerator


class NormalRule(AbstractRule):
    def __init__(self):
        super().__init__()
        self._properties = None

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, value):
        self._properties = value

    def run(self, item, **kwargs):
        return ""


class RecursiveRule(AbstractRule):
    def __init__(self):
        super().__init__()
        self._properties = None

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, value):
        self._properties = value

    def run(self, item, only_fields=False, depth=0, force=False):
        query = []
        deprecated = False
        for field in item.fields:
            if item.fields[field].is_deprecated:
                deprecated = item.fields[field].is_deprecated
            if field in self._properties.ignore:
                continue
            if only_fields:
                if field not in self._properties.only:
                    continue
            if item.fields[field].type.kind in ["OBJECT", "UNION"]:
                if force or depth <= self._properties.recursive_depth:
                    depth += 1
                    subquery_val = item.fields[field].type.name
                    subquery_val = self._properties.model.items[subquery_val]
                    subquery_val, deprecated_deep = self.run(subquery_val, only_fields, depth)
                    if deprecated_deep:
                        deprecated = deprecated_deep
                    depth -= 1
                    if subquery_val is None:
                        continue
                    if item.fields[field].type.kind == 'UNION':
                        for i in range(len(subquery_val)):
                            subquery_val[i] = '... on ' + subquery_val[i]
                    query.append((str(field) + ' {' + ' '.join(subquery_val) + '}'))
            else:
                if not force:
                    query.append(field)
        if not query:
            return self.run(item, only_fields, depth, force=True)
        return query, deprecated


class BasicQueryGenerator(AbstractGenerator):
    def __init__(self, normal: AbstractRule, recursive: AbstractRule, properties=None):
        super().__init__(normal, recursive)
        self._properties = properties
        self.recursive.properties = properties
        self.normal.properties = properties

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, value):
        self._properties = value
        self.recursive.properties = value
        self.normal.properties = value

    @property
    def normal(self):
        return self._normal

    @property
    def recursive(self):
        return self._recursive

    def generate(self, item, only_fields=False):
        args_deprecated = False
        for entry in item.args.values():
            args_deprecated = args_deprecated or args_tree_walk_for_deprecations(entry)
        if item.is_deprecated:
            args_deprecated = True
        if item.type.kind == 'OBJECT':
            try:
                subquery_val, deprecated = self.recursive.run(self._properties.model.items[item.type.name], only_fields)
                args_deprecated = args_deprecated or deprecated
            except RecursionError:
                raise
            return ' {' + ' '.join(subquery_val) + '}', args_deprecated
        else:
            return self.normal.run(self._properties.model.items[item.type.name]), args_deprecated


def args_tree_walk_for_deprecations(entry, depth=0):
    depth += 1
    if depth >= 4:
        return False
    deprecated = False
    if hasattr(entry, 'description'):
        description = entry.description
    elif hasattr(entry, '_description'):
        description = entry._description
    else:
        description = None
    if 'deprecated' in str(description).lower():
        return True
    if hasattr(entry, 'fields'):
        deprecates = []
        for name, field in entry.fields.items():
            if 'deprecated' in str(field.description).lower():
                return True
            deprecates.append(args_tree_walk_for_deprecations(field.type, depth))
        deprecated = deprecated or any(deprecates)
    return deprecated



from abc import ABC, abstractmethod


class GQBase(ABC):

    @property
    @abstractmethod
    def name(self):
        """Name of stored object"""

    @property
    @abstractmethod
    def kind(self):
        """Kind of object stored"""

    @abstractmethod
    def __init__(self, name, kind, description):
        self._name = name
        self._kind = kind
        self._description = description
        """Initialization with base params"""

    @abstractmethod
    def __repr__(self):
        """Representation of object"""

    @abstractmethod
    def parse(self, item):
        """Parsing an object"""



from gqla.abstracts import AbstractModel, GQBase


class GQModel(AbstractModel):

    def __init__(self):
        self._items = {}

    @property
    def items(self):
        return self._items

    def add_item(self, object_inst: GQBase):
        if object_inst is not None:
            self._items[object_inst.name] = object_inst



class VerticalGeneratorProperties:
    def __init__(self):
        self._ignore = None
        self._only = None
        self._model = None
        self._recursive_depth = None

    @property
    def recursive_depth(self):
        return self._recursive_depth

    @recursive_depth.setter
    def recursive_depth(self, value):
        self._recursive_depth = value

    @property
    def ignore(self):
        return self._ignore

    @ignore.setter
    def ignore(self, value):
        self._ignore = value

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value):
        self._model = value

    @property
    def only(self):
        return self._only

    @only.setter
    def only(self, value):
        self._only = value


class VerticalUrlProperties:
    def __init__(self):
        self._url = None
        self._port = None
        self._raw_query = None
        self._template = None
        self._introspection = None

    @property
    def url_string(self):
        return self.template.format(self.url, self.port)

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, value):
        self._port = value

    @property
    def template(self):
        return self._template

    @template.setter
    def template(self, value):
        self._template = value

    @property
    def raw_query(self):
        return self._raw_query

    @raw_query.setter
    def raw_query(self, value):
        self._raw_query = value

    @property
    def introspection(self):
        return self._introspection

    @introspection.setter
    def introspection(self, value):
        self._introspection = value



from gqla.GQQuery.Generator import RecursiveRule, NormalRule, BasicQueryGenerator  # nolint
from gqla.GQQuery.GQQuery import BasicQuery  # nolint

__all__ = [
    BasicQuery,
    RecursiveRule,
    NormalRule,
    BasicQueryGenerator
]



import os

try:
    with open('../VERSION', 'r') as v_file:
        VERSION = v_file.readline()
except FileNotFoundError or FileExistsError:
    VERSION = 'UNDEFINED'

LOG_LEVEL_PROJECT = os.getenv('LOG_LEVEL_PROJECT', 'DEBUG')
LOG_FORMAT_PROJECT = os.getenv('LOG_FORMAT_PROJECT',
                               '%(levelname)-8s# %(filename)s[LINE:%(lineno)d] [%(asctime)s.%(msecs)d]:  %(message)s')

LOGGING_BASE_CONFIG = {
    'version': 1,
    'formatters': {
        'default': {
            'format': LOG_FORMAT_PROJECT
        }
    },
    'handlers': {
        'console': {
            'level': LOG_LEVEL_PROJECT,
            'class': 'logging.StreamHandler',
            'formatter': 'default',
            'stream': 'ext://sys.stdout'
        },
        'error': {
            'level': 'ERROR',
            'class': 'logging.StreamHandler',
            'formatter': 'default',
            'stream': 'ext://sys.stderr'
        },
    },
    'loggings': {
        'default': {
            'level': 'DEBUG',
            'handlers': ['console', 'error']
        }
    },
    'root': {
        'level': 'DEBUG',
        'handlers': ['console', 'error']
    },
    'disable_existing_loggers': False
}



INTROSPECTION = {
    'query': '\n    query IntrospectionQuery {\n      __schema {\n        queryType { name }\n        '
             'mutationType { name }\n        subscriptionType { name }\n        types {\n          ...FullType\n  '
             '      }\n        directives {\n          name\n          description\n          locations\n         '
             ' args {\n            ...InputValue\n          }\n        }\n      }\n    }\n\n    fragment FullType '
             'on __Type {\n      kind\n      name\n      description\n      fields(includeDeprecated: true) {\n   '
             '     name\n        description\n        args {\n          ...InputValue\n        }\n        type {'
             '\n          ...TypeRef\n        }\n        isDeprecated\n        deprecationReason\n      }\n      '
             'inputFields {\n        ...InputValue\n      }\n      interfaces {\n        ...TypeRef\n      }\n    '
             '  enumValues(includeDeprecated: true) {\n        name\n        description\n        isDeprecated\n  '
             '      deprecationReason\n      }\n      possibleTypes {\n        ...TypeRef\n      }\n    }\n\n    '
             'fragment InputValue on __InputValue {\n      name\n      description\n      type { ...TypeRef }\n   '
             '   defaultValue\n    }\n\n    fragment TypeRef on __Type {\n      kind\n      name\n      ofType {'
             '\n        kind\n        name\n        ofType {\n          kind\n          name\n          ofType {'
             '\n            kind\n            name\n            ofType {\n              kind\n              '
             'name\n              ofType {\n                kind\n                name\n                ofType {'
             '\n                  kind\n                  name\n                  ofType {\n                    '
             'kind\n                    name\n                  }\n                }\n              }\n           '
             ' }\n          }\n        }\n      }\n    }\n  ',
    'variables': {}, 'operationName': None}

RAW = "query {{ {query} }}"
BASE_TEMPLATE = "http://{}:{}/graphql"



from abc import ABC, abstractmethod


class AbstractGenerator(ABC):

    @abstractmethod
    def __init__(self, normal, recursive):
        self._normal = normal
        self._recursive = recursive

    @property
    @abstractmethod
    def normal(self):
        """Normal rule"""

    @property
    @abstractmethod
    def recursive(self):
        """Recursive rule"""

    @abstractmethod
    def generate(self, *args):
        """Generate statement"""


class AbstractRule(ABC):

    @abstractmethod
    def run(self, *args):
        """Rule runner"""



from abc import ABC, abstractmethod

from gqla.abstracts.GQLStorage import GQBase


class AbstractModel(ABC):

    @property
    @abstractmethod
    def items(self):
        """Item storage"""

    @abstractmethod
    def add_item(self, object_inst: GQBase):
        """Add item to storage"""



from abc import ABC, abstractmethod


class AbstractStorage(ABC):
    @property
    @abstractmethod
    def storage(self):
        """Query to return from storage"""

    @abstractmethod
    def add(self, *args):
        """Add to storage"""

    @abstractmethod
    def create(self, *args):
        """Set query value"""



import setuptools

with open('VERSION', 'r') as verfile:
    version = verfile.read()

with open("README.md", 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name="gqla",
    version=version,
    author="Alexey Kuzin",
    author_email="alenstoir@yandex.ru",
    description="A module used to generate querry statements and perform data fetching via GraphQL",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Alenstoir/GQLA",
    license='MIT License',
    requires='aiohttp',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
)



from gqla.abstracts import GQBase


class GQField:
    def __init__(self, type_, is_deprecated=False, deprecation_reason=None, description=False, args=None):
        if args is None:
            args = {}
        self.type = type_
        self.is_deprecated = is_deprecated
        self.deprecation_reason = deprecation_reason
        self.description = description
        self.args = {}
        for arg in args:
            self.args[arg.name] = arg


class GQENUM(GQBase):

    def __init__(self, name, kind, description=None, values=None):
        super().__init__(name, kind, description)
        self.values = values

    @property
    def name(self):
        return self._name

    @property
    def kind(self):
        return self._kind

    def __repr__(self):
        answer = ','.join(['name: ' + self.name, ' kind:' + self.kind, ' values:' + str(self.values)])
        return answer

    def update(self, item, *args):
        values = []
        if 'enumValues' in item:
            for enum in item['enumValues']:
                if enum['name'] not in self.values:
                    values.append(enum['name'])
        self.values = values
        return self

    def parse(self, item, model=None):
        values = []
        if 'enumValues' in item:
            for enum in item['enumValues']:
                values.append(enum['name'])
        self.values = values
        return self


class GQUNION(GQBase):

    def __init__(self, name, kind, description=None):
        super().__init__(name, kind, description)
        self._fields = {}

    def add_field(self, name, field: GQField):
        self._fields[name] = field

    @property
    def name(self):
        return self._name

    @property
    def kind(self):
        return self._kind

    @property
    def fields(self):
        return self._fields

    def __repr__(self):
        answer = ','.join(['name: ' + self.name, ' kind:' + self.kind, ' fields:['])
        for field in self.fields.values():
            answer += '... on ' + field.type.name + ' {' + str(field.type) + '},'
        answer = answer.strip(',') + ']'
        return answer
        pass

    def update(self, item, model=None):
        if 'possibleTypes' in item:
            for object_ in item['possibleTypes']:
                if object_['name'] in self.fields:
                    self.fields[object_['name']].type.update(object_, model)
                else:
                    obj = TypeFactory(object_, model)
                    self.add_field(object_['name'], GQField(obj.parse(object_, model)))
        return self

    def parse(self, item, model=None):
        if 'possibleTypes' in item:
            for object_ in item['possibleTypes']:
                obj = TypeFactory(object_, model)
                self.add_field(object_['name'], GQField(obj.parse(object_, model), object_['isDeprecated'],
                                                        object_['deprecationReason']))
        return self


class GQJSON(GQBase):

    def __init__(self, name, kind, description=None):
        super().__init__(name, kind, description)

    @property
    def name(self):
        return self._name

    @property
    def kind(self):
        return self._kind

    def __repr__(self):
        pass

    def update(self, item, model=None):
        return self

    def parse(self, item, model=None):
        return self


class GQSCALAR(GQBase):

    def __init__(self, name, kind, description=None):
        super().__init__(name, kind, description)

    @property
    def name(self):
        return self._name

    @property
    def kind(self):
        return self._kind

    def __repr__(self):
        answer = ','.join(['name: ' + self.name, ' kind:' + self.kind])
        return answer

    def update(self, item, model=None):
        self._description = item.get('description')
        return self

    def parse(self, item, model=None):
        return self


class GQOBJECT(GQBase):

    @property
    def name(self):
        return self._name

    @property
    def kind(self):
        return self._kind

    def __init__(self, name, kind, description=None):
        super().__init__(name, kind, description)
        self._fields = {}

    @property
    def fields(self):
        return self._fields

    def add_field(self, name, field: GQField):
        self._fields[name] = field

    def __repr__(self):
        answer = ','.join(['name: ' + self.name, ' kind:' + self.kind, ' fields:['])
        for field in self.fields.values():
            answer += '{' + str(field) + '},'
        answer = answer.strip(',') + ']'
        return answer
        pass

    def update(self, item, model=None):
        if 'fields' in item:
            for field in item['fields']:
                obj = TypeFactory(field, model)
                if obj is not None:
                    if field['args']:
                        args = [TypeFactory(arg, model) for arg in field['args']]
                    else:
                        args = {}
                    if field['name'] not in self.fields:
                        self.add_field(field['name'], GQField(obj, field['isDeprecated'], field['deprecationReason'],
                                                              field['description'], args))
                    else:
                        self.fields[field['name']].type.update(field, model)
        return self

    def parse(self, item, model=None):
        if 'fields' in item:
            for field in item['fields']:
                obj = TypeFactory(field, model)
                if obj is not None:
                    if field['args']:
                        args = [TypeFactory(arg, model) for arg in field['args']]
                    else:
                        args = {}
                    self.add_field(field['name'],
                                   GQField(obj, field['isDeprecated'], field['deprecationReason'], field['description'],
                                           args))
        return self


class GQINPUT_OBJECT(GQOBJECT):
    def update(self, item, model=None):
        if 'inputFields' in item:
            for field in item['inputFields']:
                obj = TypeFactory(field, model)
                if obj is not None:
                    if field['name'] not in self.fields:
                        self.add_field(field['name'], GQField(obj, description=field['description']))
                    else:
                        self.fields[field['name']].type.update(field, model)
        return self

    def parse(self, item, model=None):
        if 'inputFields' in item:
            for field in item['inputFields']:
                obj = TypeFactory(field, model)
                if obj is not None:
                    self.add_field(field['name'], GQField(obj, description=field['description']))
        return self


def TypeFactory(kind, model=None):  # noqa
    if 'type' in kind:
        item = kind['type']
    else:
        item = kind
    while True:
        if item['name'] is None:
            item = item['ofType']
        else:
            class_name = item['kind']
            item_name = item['name']
            break
    possibles = globals().copy()
    possibles.update(locals())
    class_instance = possibles.get('GQ' + class_name)
    if not class_instance:
        return None
    if item_name not in model.items:
        obj = class_instance(item_name, class_name, kind.get('description'))
        model.add_item(obj.parse(kind, model))
    else:
        obj = model.items[item_name]
        if kind.get('kind', '') != 'SCALAR' and kind['name'] != item_name:
            return obj
        obj.update(kind, model)
    return obj



import asyncio
import json
import logging
import logging.config
import os.path

from gqla.Executor import BasicExecutor, AsyncRunner
from gqla.GQLModel import GQModel
from gqla.GQLStorage import TypeFactory
from gqla.GQQStorage import BasicStorage
from gqla.GQQuery import NormalRule, RecursiveRule, BasicQueryGenerator
from gqla.VerticalStorage import VerticalGeneratorProperties, VerticalUrlProperties
from gqla.abstracts import AbstractStorage, AbstractExecutor
from gqla.statics import INTROSPECTION, BASE_TEMPLATE, RAW


class GQLA:
    __slots__ = ('name', '_subpid', 'usefolder', '_depth', '_folder', '_pretty', '_qStorage', '_executor',
                 '_gen_properties', '_url_properties', '_default_normal_rule', '_default_recursive_rule',
                 '_default_gen', '_basic_runner')

    def __init__(self, name, url=None, port=None, ignore=None, only=None, usefolder=False, recursive_depth=5):
        super().__init__()

        self._executor = None
        self._qStorage = None

        self._basic_runner = AsyncRunner()
        self._gen_properties = VerticalGeneratorProperties()
        self._url_properties = VerticalUrlProperties()
        self._default_normal_rule = NormalRule()
        self._default_recursive_rule = RecursiveRule()
        self._default_gen = BasicQueryGenerator(self._default_normal_rule, self._default_recursive_rule)
        self.qStorage = BasicStorage(self._gen_properties, self._default_gen)
        self.executor = BasicExecutor(self._url_properties, storage=self.qStorage, runner=self._basic_runner)

        self._gen_properties.model = GQModel()
        self._gen_properties.ignore = ignore
        self._gen_properties.only = only
        self._gen_properties.recursive_depth = recursive_depth

        self._url_properties.url = url
        self._url_properties.port = port
        self._url_properties.template = BASE_TEMPLATE
        self._url_properties.introspection = INTROSPECTION
        self._url_properties.raw_query = RAW

        self.usefolder = usefolder
        self._pretty = False
        self.name = name
        self._subpid = 0

        logging.info(' '.join(['CREATED', 'CLASS', str(self.__class__)]))

        if self.usefolder:
            self._folder = os.path.join('', self.name)
            if not os.path.exists(self._folder):
                os.mkdir(self._folder)

    # Getters\Setters

    @property
    def executor(self):
        return self._executor

    @executor.setter
    def executor(self, value: AbstractExecutor):
        self._executor = value

    @property
    def url(self):
        return self._url_properties.url

    @url.setter
    def url(self, value):
        self._url_properties.url = value

    @property
    def port(self):
        return self._url_properties.port

    @port.setter
    def port(self, value):
        self._url_properties.port = value

    @property
    def qStorage(self):
        return self._qStorage

    @qStorage.setter
    def qStorage(self, value: AbstractStorage):
        self._qStorage = value
        if self.executor is not None:
            self.executor.storage = value

    @property
    def ignore(self):
        return self._gen_properties.ignore

    @ignore.setter
    def ignore(self, value):
        self._gen_properties.ignore = value

    @property
    def only(self):
        return self._gen_properties.only

    @only.setter
    def only(self, value):
        self._gen_properties.only = value

    @property
    def recursive_depth(self):
        return self._gen_properties.recursive_depth

    @recursive_depth.setter
    def recursive_depth(self, value):
        self._gen_properties.recursive_depth = value

    @property
    def model(self):
        return self._gen_properties.model

    @model.setter
    def model(self, value):
        self._gen_properties.model = value

    # Getters\Setters

    async def query_one(self, query_name, only_fields=False, usefolder=False, **kwargs):
        query = query_name
        if isinstance(query_name, dict):
            query_name = query_name['query'].split('{')[0].strip(' \n').split(' ')[1]
        if only_fields:
            self.generate_queries(specific=query_name, only_fields=only_fields)
            query_name = query = "only_" + query_name
        logging.info("FETCHING " + query_name + " WITH ARGS " + str(kwargs))
        result = await self.executor.execute(self._subpid, query, **kwargs)
        self._subpid += 1

        if self.usefolder:
            if usefolder:
                filename = os.path.join(self._folder, query_name + '.json')
                logging.info(' '.join(['WRITING', query_name, 'RESULT TO', filename]))
                with open(filename, 'w') as ofs:
                    if self._pretty:
                        indent = 4
                    else:
                        indent = None
                    ofs.write(json.dumps(result, indent=indent))
        return result

    async def introspection(self):
        logging.info(' '.join(['QUERRYING', self.name, 'INTROSPECTION']))
        result = await self.query_one(INTROSPECTION, usefolder=True)
        queries = result['data']['__schema']['types']

        self.create_data(queries)
        self.generate_queries()

    def create_data(self, data):
        for item in data:
            TypeFactory(item, self.model)

    def generate_queries(self, specific=False, only_fields=False):
        if 'Query' in self.model.items:
            queries = self.model.items['Query'].fields
        elif 'Queries' in self.model.items:
            queries = self.model.items['Queries'].fields
        else:
            raise NotImplementedError
        if not specific:
            for query in queries:
                self.qStorage.create(query, query, queries[query], self.recursive_depth)
        else:
            for query in queries:
                if query == specific:
                    query_name = query if not only_fields else "only_" + query
                    self.qStorage.create(query_name, query, queries[query], self.recursive_depth, only_fields)

        self.executor._storage = self.qStorage


"""
This file defines main class which is used to generate and query GraphQL queries
All what lies bellow is standalone support 
"""


async def asynchronous():  # Пример работы
    ignore = ['pageInfo', 'deprecationReason', 'isDeprecated', 'cursor', 'parent1', 'id']
    only = ['edges', 'node', 'code', 'name', 'StarObject', 'PlanetObject', 'orbitals']

    helper = GQLA('solar', url='localhost', port='8080', usefolder=True, ignore=ignore, recursive_depth=5)
    helper.only = only
    helper._pretty = True
    await helper.introspection()
    
    result = await helper.query_one('allStellar', usefolder=True, filters={'not': {'objectType': 'planet'}}, first='5')
    print(result)

    result = await helper.query_one('allStellar', usefolder=False, only_fields=True, first='1')
    print(result)

if __name__ == "__main__":
    from gqla.settings import LOGGING_BASE_CONFIG

    logging.getLogger(__name__)
    logging.config.dictConfig(LOGGING_BASE_CONFIG)
    loop_ = asyncio.get_event_loop()
    loop_.run_until_complete(asynchronous())

    loop_.close()
    pass



from gqla.abstracts.Executor import AbstractRunner, AbstractExecutor
from gqla.abstracts.GQGenerator import AbstractRule, AbstractGenerator
from gqla.abstracts.GQLModel import AbstractModel
from gqla.abstracts.GQLStorage import GQBase
from gqla.abstracts.GQQStorage import AbstractStorage
from gqla.abstracts.GQQery import AbstractQuery

__all__ = [
    GQBase,
    AbstractStorage,
    AbstractRule,
    AbstractGenerator,
    AbstractQuery,
    AbstractRunner,
    AbstractExecutor
]
